<?php
/**
 * Copyright 2018 Kacper Donat
 *
 * @author Kacper "Kadet" Donat <kacper@kadet.net>
 *
 * Full license available in separate LICENSE file
 */

namespace Kadet\Functional\Utils;

use const Kadet\Functional\_ as PLACEHOLDER;
use Kadet\Functional\Decorator;
use function Kadet\Functional\not;
use function Kadet\Functional\Predicates\same;

class PartiallyAppliedFunction implements Decorator
{
    protected $callable;
    protected $arguments = [];

    public function __construct(callable $callable, ...$args)
    {
        switch (true) {
            case $callable instanceof PartiallyAppliedFunction:
                $this->callable  = $callable->getDecorated();
                $this->arguments = $callable->prepareArguments($args);
                break;

            default:
                $this->callable  = $callable;
                $this->arguments = array_values($args);
        }
    }

    public function getArguments()
    {
        return array_filter($this->arguments, not(same(PLACEHOLDER)));
    }

    public function __invoke(...$args)
    {
        $arguments = $this->prepareArguments($args);
        return ($this->callable)(...$arguments);
    }

    protected function prepareArguments($args)
    {
        $result = [];
        foreach ($this->arguments as $argument) {
            $result[] = $argument === PLACEHOLDER && !empty($args)
                ? array_shift($args)
                : $argument;
        }

        return array_merge(
            array_values($result),
            array_values($args)
        );
    }

    public function getDecorated()
    {
        return $this->callable;
    }

    public static function fix(callable $callable, ...$args)
    {
        $applied = new self($callable, ...$args);
        array_unshift($applied->arguments, $applied);

        return $applied;
    }
}